@@ -2,7 +2,7 @@ use anyhow::{anyhow, bail, Result};
 use clap::Parser;
 use cli::{Opts, SubCommands};
 use core::default::Default;
-use dialoguer::{Confirm, Select};
+use dialoguer::{Confirm, MultiSelect};
 use serde::{Deserialize, Serialize};
 use std::{
     fs,
@@ -11,16 +11,17 @@ use std::{
 };
 
 use i18n::I18N_LOADER;
-use kernel::Kernel;
+use kernel::{generic_kernel::GenericKernel, Kernel};
 
 mod cli;
 mod i18n;
 mod kernel;
 mod macros;
-mod parser;
+mod version;
 
 const CONF_PATH: &str = "/etc/systemd-boot-friend.conf";
 const REL_DEST_PATH: &str = "EFI/systemd-boot-friend/";
+const SRC_PATH: &str = "/boot";
 
 #[derive(Debug, Serialize, Deserialize)]
 pub struct Config {
@@ -49,51 +50,55 @@ impl Default for Config {
 }
 
 /// Choose a kernel using dialoguer
-fn choose_kernel(kernels: &[Kernel], prompt: String) -> Result<Kernel> {
+fn choose_kernel<K: Kernel>(kernels: &[K], prompt: &str) -> Result<Vec<K>> {
     if kernels.is_empty() {
         bail!(fl!("empty_list"));
     }
 
-    // build dialoguer Select for kernel selection
-    let n = Select::new()
+    // build dialoguer MultiSelect for kernel selection
+    Ok(MultiSelect::new()
         .with_prompt(prompt)
         .items(kernels)
-        .default(0)
-        .interact()?;
+        .interact()?
+        .iter()
+        .map(|n| kernels[*n].clone())
+        .collect())
+}
 
-    Ok(kernels[n].clone())
+#[inline]
+fn parse_num_or_filename(
+    config: &Config,
+    n: &str,
+    kernels: &[GenericKernel],
+) -> Result<GenericKernel> {
+    match n.parse::<usize>() {
+        Ok(num) => Ok(kernels
+            .get(num - 1)
+            .ok_or_else(|| anyhow!(fl!("invalid_index")))?
+            .clone()),
+        Err(_) => GenericKernel::parse(config, n),
+    }
 }
 
 #[inline]
 fn specify_or_choose(
     config: &Config,
-    arg: Option<String>,
-    kernels: &[Kernel],
-    prompt: String,
-) -> Result<Kernel> {
+    arg: &Option<String>,
+    kernels: &[GenericKernel],
+    prompt: &str,
+) -> Result<Vec<GenericKernel>> {
     match arg {
         // the target can be both the number in
         // the list and the name of the kernel
-        Some(n) => parse_num_or_filename(config, &n, kernels),
+        Some(n) => Ok(vec![parse_num_or_filename(config, n, kernels)?]),
         // select the kernel to remove
         // when no target is given
         None => choose_kernel(kernels, prompt),
     }
 }
 
-#[inline]
-fn parse_num_or_filename(config: &Config, n: &str, kernels: &[Kernel]) -> Result<Kernel> {
-    match n.parse::<usize>() {
-        Ok(num) => Ok(kernels
-            .get(num - 1)
-            .ok_or_else(|| anyhow!(fl!("invalid_index")))?
-            .clone()),
-        Err(_) => Kernel::parse(config, n),
-    }
-}
-
 /// Initialize the default environment for friend
-fn init(config: &Config, installed_kernels: &[Kernel], kernels: &[Kernel]) -> Result<()> {
+fn init<K: Kernel>(config: &Config, installed_kernels: &[K], kernels: &[K]) -> Result<()> {
     // use bootctl to install systemd-boot
     println_with_prefix_and_fl!("init");
     print_block_with_fl!("prompt_init");
@@ -103,7 +108,7 @@ fn init(config: &Config, installed_kernels: &[Kernel], kernels: &[Kernel]) -> Re
         .default(false)
         .interact()?
     {
-        Command::new("bootctl")
+        let child_output = Command::new("bootctl")
             .arg("install")
             .arg(
                 "--esp=".to_owned()
@@ -112,15 +117,20 @@ fn init(config: &Config, installed_kernels: &[Kernel], kernels: &[Kernel]) -> Re
                         .to_str()
                         .ok_or_else(|| anyhow!(fl!("invalid_esp")))?,
             )
-            .stderr(Stdio::null())
-            .spawn()?;
+            .stderr(Stdio::piped())
+            .spawn()?
+            .wait_with_output()?;
+
+        if !child_output.status.success() {
+            bail!(String::from_utf8(child_output.stderr)?);
+        }
 
         // create folder structure
         println_with_prefix_and_fl!("create_folder");
         fs::create_dir_all(config.esp_mountpoint.join(REL_DEST_PATH))?;
 
         // Update systemd-boot kernels and entries
-        print_block_with_fl!("prompt_update");
+        print_block_with_fl!("prompt_update", src_path = SRC_PATH);
         Confirm::new()
             .with_prompt(fl!("ask_update"))
             .default(false)
@@ -133,8 +143,9 @@ fn init(config: &Config, installed_kernels: &[Kernel], kernels: &[Kernel]) -> Re
 }
 
 /// Update systemd-boot kernels and entries
-fn update(installed_kernels: &[Kernel], kernels: &[Kernel]) -> Result<()> {
+fn update<K: Kernel>(installed_kernels: &[K], kernels: &[K]) -> Result<()> {
     println_with_prefix_and_fl!("update");
+    print_block_with_fl!("note_copy_files");
 
     // Remove existing kernels
     installed_kernels.iter().try_for_each(|k| k.remove())?;
@@ -153,7 +164,9 @@ fn update(installed_kernels: &[Kernel], kernels: &[Kernel]) -> Result<()> {
 }
 
 #[inline]
-fn install(kernel: &Kernel, force: bool) -> Result<()> {
+fn install<K: Kernel>(kernel: &K, force: bool) -> Result<()> {
+    print_block_with_fl!("note_copy_files");
+
     kernel.install_and_make_config(force)?;
     kernel.ask_set_default()?;
 
@@ -161,7 +174,7 @@ fn install(kernel: &Kernel, force: bool) -> Result<()> {
 }
 
 #[inline]
-fn print_kernels(kernels: &[Kernel]) {
+fn print_kernels<K: Kernel>(kernels: &[K]) {
     kernels
         .iter()
         .enumerate()
@@ -190,7 +203,7 @@ fn read_config() -> Result<Config> {
             println_with_prefix_and_fl!("conf_default", conf_path = CONF_PATH);
             fs::create_dir_all(PathBuf::from(CONF_PATH).parent().unwrap())?;
             fs::write(CONF_PATH, toml::to_string_pretty(&Config::default())?)?;
-            bail!(fl!("edit_conf", conf_path = CONF_PATH))
+            Err(anyhow!(fl!("edit_conf", conf_path = CONF_PATH)))
         }
     }
 }
@@ -201,25 +214,27 @@ fn main() -> Result<()> {
 
     // Read config, create a default one if the file is missing
     let config = read_config()?;
-    let installed_kernels = Kernel::list_installed(&config)?;
-    let kernels = Kernel::list(&config)?;
+    let installed_kernels = GenericKernel::list_installed(&config)?;
+    let kernels = GenericKernel::list(&config)?;
 
     // Switch table
     match matches.subcommands {
         Some(s) => match s {
             SubCommands::Init => init(&config, &installed_kernels, &kernels)?,
             SubCommands::Update => update(&installed_kernels, &kernels)?,
-            SubCommands::InstallKernel(args) => install(
-                &specify_or_choose(&config, args.target, &kernels, fl!("select_install"))?,
-                args.force,
-            )?,
+            SubCommands::InstallKernel(args) => {
+                specify_or_choose(&config, &args.target, &kernels, &fl!("select_install"))?
+                    .iter()
+                    .try_for_each(|k| install(k, args.force))?
+            }
             SubCommands::RemoveKernel(args) => specify_or_choose(
                 &config,
-                args.target,
+                &args.target,
                 &installed_kernels,
-                fl!("select_remove"),
+                &fl!("select_remove"),
             )?
-            .remove()?,
+            .iter()
+            .try_for_each(|k| k.remove())?,
             SubCommands::ListAvailable => print_kernels(&kernels),
             SubCommands::ListInstalled => print_kernels(&installed_kernels),
         },
